void sphereFold(float3* z, float* dz)
{
	const float minRadius2 = 0.25f;
	const float fixedRadius2 = 0.1f;

	float r2 = dot(*z,*z);
	if (r2<minRadius2){ 
		// linear inner scaling
		float temp = (fixedRadius2/minRadius2);
		*z *= temp;
		*dz*= temp;
	}
	else if (r2<fixedRadius2){ 
		// this is the actual sphere inversion
		float temp =(fixedRadius2/r2);
		*z *= temp;
		*dz*= temp;
	}
}

void boxFold(float3* z, float* dz) 
{
	//const float foldingLimit = 0.25f;
	const float foldingLimit = 1.0f;
	
	*z = clamp(*z, -foldingLimit, foldingLimit) * 2.0 - *z;
}
float de_mandelbox(float3 z)
{
	const int Iterations = 20;
	//const float Scale = 1.60f;
	const float Scale = -2.6f;

	float3 offset = z;
	float dr = 1.0;
	for (int n = 0; n < Iterations; n++){
		boxFold(&z,&dr);       // Reflect
		sphereFold(&z,&dr);    // Sphere Inversion
 		
        z=Scale*z + offset;  // Scale & Translate
        dr = dr*fabs(Scale)+1.0;
	}
	float r = length(z);
	return r/fabs(dr);
}

float de_mandelbulb(float3 pos)
{
	const int Iterations = 8;
	const float Bailout = 2.0;
	const float Power = 8.0;
	float anim = ANIM;
	//float Power = 5.5 + 3.0 * sin(anim*6.28/72.0);

	float3 z = pos;
	float dr = 1.0;
	float r = 0.0;
	for (int i = 0; i < Iterations ; i++) {
		r = length(z);
		if (r>Bailout) break;
		
		// convert to polar coordinates
		float theta = acos(z.z/r);
		float phi = atan2(z.y,z.x);
		dr = pow(r, Power - (float)1.0) * Power*dr + 1.0;
		
		// scale and rotate the point
		float zr = pow( r,Power);
		theta = theta*Power;
		phi = phi*Power;
		
		// convert back to cartesian coordinates
		z = zr*(float3)(sin(theta)*cos(phi), sin(phi)*sin(theta), cos(theta));
		z+=pos;
	}
	return 0.5*log(r)*r/dr;
}

float de_sierp(float3 z)
{
	const int Iterations = 32;
	const float Scale = 2.0;
	const float Factor = 1.0;

	float3 a1 = Factor*(float3)(1,1,1);
	float3 a2 = Factor*(float3)(-1,-1,1);
	float3 a3 = Factor*(float3)(1,-1,-1);
	float3 a4 = Factor*(float3)(-1,1,-1);
	float3 c;
	int n = 0;
	float dist, d;
	while (n < Iterations) {
		c = a1;
		dist = length(z-a1);
	    d = length(z-a2);
		if (d < dist) { c = a2; dist=d; }
		d = length(z-a3);
		if (d < dist) { c = a3; dist=d; }
		d = length(z-a4);
		if (d < dist) { c = a4; dist=d; }
		z = Scale*z-c*(Scale-1.0);
		n++;
	}
	return length(z) * pow(Scale, (float)(-n));
}

float3 bg_rotate(float3 v, float x, float y, float z)
{
	float sinx = sin(x);
	float cosx = cos(x);
	float siny = sin(y);
	float cosy = cos(y);
	
	float3 roty = (float3)(.0,.0,.0);
	roty.x = v.x * cosy + v.z * siny;
	roty.y = v.y;
	roty.z = -v.x * siny + v.z * cosy;
	float3 rotx = (float3)(.0,.0,.0);
	rotx.x = roty.x;
	rotx.y = roty.y * cosx - roty.z * sinx;
	rotx.z = roty.y * sinx + roty.z * cosx;
	return rotx;
}

kernel void background(global float4* dir, 
					global RaytraceInfo* work,
                    //global float4* colorOut,
					int recursion)
{
	float anim = ANIM;
    int x = get_global_id(0);
	RaytraceInfo rtmp = work[x];
	if(rtmp.state == 0){
		float3 ddir = normalize(dir[x].xyz);
		//ddir = normalize(bg_rotate(ddir,.15,5.0*6.28/72.0,10.0));  //use bg_rotate to rotate the direction
	
		float3 pos = -1.2 * ddir;  //use direction as a position to calculate from
		int iterations = 0;
		float de = de_mandelbulb(pos);   //change inside loop too!
		//float de = de_mandelbox(pos);      //change inside loop too!
		//float de = de_sierp(pos);        //change inside loop too!
		float totdistance = de;
		float precision = 0.00001;
		while((de > precision) && (iterations < 320)){
			pos += ddir * de;
			de = de_mandelbulb(pos);     //change outside loop too!
			//de = de_mandelbox(pos);        //change outside loop too!
			//de = de_sierp(pos);          //change outside loop too!
			totdistance += de;
			iterations++;
		}
		
		float cdist = totdistance / 4.0;
		float citer = (float)iterations / 32.0;
		float blue = 1.0 - clamp(cdist * 0.2 + citer * 1.5, 0.0, 1.0);
		float green = 1.0 - clamp(sin(cdist * 8.0) * 0.8 + citer * 0.2, 0.0, 1.0);
		float red = 1.0 - clamp(cdist * 0.8 + citer * 0.2, 0.0, 1.0);
		
		work[x].sphere_color = (float4)(red, green, blue, .0);
	}
}

